require 'rbconfig'
require 'tempfile'
require 'stringio'

module RutokenMgr
  module UI
    class MainApp < GladeBase
      include Singleton
      include GetText
      GetText.bindtextdomain(RutokenMgr::TEXTDOMAIN, nil, nil, "UTF-8")

      attr_accessor :main_window, :prefs
      def initialize
        super "main_app.glade"
        @prefs = Preferences.instance
        setup_data
        @main_window.show_all

        @prefs.sopin = PINDialog.new(_("Security Officer PIN"), @main_window, '').execute
        exit if @prefs.sopin.nil?
      end

      def setup_data
        @main_window.title = TITLE

        @card_select.model = Gtk::ListStore.new(Integer, String)
        renderer = Gtk::CellRendererText.new
        @card_select.pack_start(renderer, true)
        @card_select.set_attributes(renderer, :text => 1)
        @card_select.active = 0

        @key_algo_select.model = Gtk::ListStore.new(String, String)
        [["rsa", _("RSA")], ["dsa", _("DSA")]].each do |k,v|
          row = @key_algo_select.model.append
          row[0] = k
          row[1] = v
        end
        renderer = Gtk::CellRendererText.new
        @key_algo_select.pack_start(renderer, true)
        @key_algo_select.set_attributes(renderer, :text => 1)
        @key_algo_select.active = 0
        @key_algo_select.sensitive = false

        @bits_entry.text = "1024"
        @bits_entry.sensitive = false

        refresh_cards
        update_checkboxen
      end

      def quit
        @main_window.destroy
      end

      def gtk_main_quit
        Gtk.main_quit
      end

      def update_checkboxen
        if (@erase_card_check.active? or
            @create_pkcs11_check.active? or
            @gen_key_check.active? or
            @gen_csr_check.active?)
            @upload_cert_check.active = false
            @upload_cert_check.sensitive = false
        else
          @upload_cert_check.sensitive = true
        end

        if @create_pkcs11_check.active?
          @pin_label_entry.sensitive = true
        else
          @pin_label_entry.sensitive = false
        end

        if @gen_csr_check.active?
          @cn_entry.sensitive = true
          @email_entry.sensitive = true
          @ou_entry.sensitive = true
          @o_entry.sensitive = true
          @c_entry.sensitive = true
          @self_sign_check.sensitive = true
        else
          @cn_entry.sensitive = false
          @email_entry.sensitive = false
          @ou_entry.sensitive = false
          @o_entry.sensitive = false
          @c_entry.sensitive = false
          @self_sign_check.active = false
          @self_sign_check.sensitive = false
        end

        if @upload_cert_check.active?
          @erase_card_check.active = false
          @erase_card_check.sensitive = false
          @create_pkcs11_check.active = false
          @create_pkcs11_check.sensitive = false
          @gen_key_check.active = false
          @gen_key_check.sensitive = false
          @gen_csr_check.active = false
          @gen_csr_check.sensitive = false
          @cn_entry.sensitive = false
          @email_entry.sensitive = false
          @ou_entry.sensitive = false
          @o_entry.sensitive = false
          @c_entry.sensitive = false
          @self_sign_check.active = false
          @self_sign_check.sensitive = false
        else
          @erase_card_check.sensitive = true
          @create_pkcs11_check.sensitive = true
          @gen_key_check.sensitive = true
          @gen_csr_check.sensitive = true
        end

        if ((not @erase_card_check.active? and
             not @create_pkcs11_check.active? and
             not @gen_key_check.active? and
             not @gen_csr_check.active? and
             not @upload_cert_check.active?) or
             (@create_pkcs11_check.active? and @pin_label_entry.text.empty?) or
             (@gen_csr_check.active? and (@cn_entry.text.empty? or @email_entry.text.empty?)))
             @ok_button.sensitive = false
        else
          @ok_button.sensitive = true
        end
      end

      def refresh_cards
        @card_select.model.clear
        IO.popen ["opensc-tool", "--list-readers"] do |opensc|
          opensc.each_line do |l|
            next unless l =~ /^\d/
            id, driver, name = l.chomp.split(/\s+/, 3)
            row = @card_select.model.append
            row[0] = id.to_i
            row[1] = name
          end
        end
        @card_select.active = 0
      end

      def system(*args)
        Tempfile.open(["output-"]) do |tmp|
          @busy_window.update @action if @busy_window
          status = Kernel.system *args, [STDOUT, STDERR] => tmp.fileno
          tmp.rewind
          output = tmp.read.chomp
          tmp.close!
          if status != true
            raise _("%s reports:\n\n%s") % [args.first, output]
          end
        end
      end

      def process_card
        userpin = nil
        cert_file = nil

        if (@create_pkcs11_check.active? or
            @gen_key_check.active? or
            @gen_csr_check.active? or
            @self_sign_check.active? or
            @upload_cert_check.active?)
          userpin = PINDialog.new(_("User PIN"), @main_window, '').execute
          return if userpin.nil?
        end

        if @gen_csr_check.active? or @upload_cert_check.active?
          chooser = Gtk::FileChooserDialog.new(
            @gen_csr_check.active? ? (@self_sign_check.active? ? _("Save certificate") : _("Save sign request")) : _("Load certificate"),
            @main_window,
            @upload_cert_check.active? ? Gtk::FileChooser::ACTION_OPEN : Gtk::FileChooser::ACTION_SAVE,
            nil,
            [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL],
            [Gtk::Stock::OPEN, Gtk::Dialog::RESPONSE_ACCEPT]
          )

          filter = Gtk::FileFilter.new
          filter.add_pattern((@self_sign_check.active? or @upload_cert_check.active?) ? "*.pem" : "*.csr")

          chooser.filter = filter

          chooser.do_overwrite_confirmation = @gen_csr_check.active?

          if chooser.run == Gtk::Dialog::RESPONSE_ACCEPT
            cert_file = chooser.filename
          end
          chooser.destroy
          return if cert_file.nil?
        end

        @busy_window = BusyWindow.new(@main_window)
        begin
          process_card!(userpin, cert_file)
        rescue => err
          alert = Gtk::MessageDialog.new(
            @main_window,
            Gtk::Dialog::MODAL,
            Gtk::MessageDialog::ERROR,
            Gtk::MessageDialog::BUTTONS_CLOSE
          )
          alert.text = _("Error while ") + @action
          alert.secondary_text = err.message

          alert.run
          alert.destroy
        ensure
          @busy_window.busy_window.destroy if @busy_window
          @busy_window = nil
        end
      end

      def process_card!(userpin, cert_file)
        @action = ""

        pkcs_init_args = [
          '--reader', @card_select.active_iter[0].to_s,
          '--so-puk', '',
          '--puk', ''
        ]
        if @erase_card_check.active?
          @action = _("erasing card")
          system 'pkcs15-init', *pkcs_init_args, '--erase-card'
        end

        if @create_pkcs11_check.active?
          # Rutoken driver only accepts "87654321" as SOPIN, "12345678" as
          # User PIN and "12345678" as "Unspecified PIN [reference 1]" during
          # initialization.
          # In Soviet Russia PINs enters YOU!
          @action = _("creating PKCS#11 structure")
          system 'pkcs15-init', *pkcs_init_args, '--create-pkcs15', '--label', @pin_label_entry.text, '--so-pin', '87654321', '--secret', '@=12345678'
          @action = _("initializing User PIN")
          system 'pkcs15-init', *pkcs_init_args, '--store-pin', '--auth-id', '02', '--label', 'User PIN', '--pin', '12345678'
          @action = _("setting SO PIN")
          system 'pkcs15-tool', '--no-cache', '--reader', @card_select.active_iter[0].to_s, '--change-pin', '--auth-id', '01', '--pin', '87654321', '--new-pin', @prefs.sopin
          @action = _("setting User PIN")
          system 'pkcs15-tool', '--no-cache', '--reader', @card_select.active_iter[0].to_s, '--change-pin', '--auth-id', '02', '--pin', '12345678', '--new-pin', userpin
        end

        # pkcs15-tool does not support --options-file, so why bother?
        pkcs_init_args << '--so-pin' << @prefs.sopin << '--pin' << userpin

        if @gen_key_check.active?
          @action = _("generating key")
          system 'pkcs15-init', *pkcs_init_args, '--generate-key', @key_algo_select.active_iter[0] + '/' + @bits_entry.text, '--auth-id', '02'
        end


        if @gen_csr_check.active?
          Tempfile.open(["openssl-", ".cnf"]) do |ossl|
            ossl.print <<-EOS
openssl_conf = openssl_def

[openssl_def]
engines = engine_section

[engine_section]
pkcs11 = pkcs11_section

[pkcs11_section]
engine_id = pkcs11
dynamic_path = #{RbConfig::CONFIG['libdir']}/openssl/engines/engine_pkcs11.so
MODULE_PATH = #{RbConfig::CONFIG['libdir']}/opensc-pkcs11.so
PIN = #{userpin}
init = 0

[req]
distinguished_name = req_distinguished_name
prompt = no

[req_distinguished_name]
EOS
            ossl.printf "CN = %s\n", @cn_entry.text
            ossl.printf "emailAddress = %s\n", @email_entry.text
            ossl.printf "OU = %s\n", @ou_entry.text unless @ou_entry.text.empty?
            ossl.printf "O = %s\n", @o_entry.text unless @o_entry.text.empty?
            ossl.printf "C = %s\n", @c_entry.text unless @c_entry.text.empty?
            ossl.close

            ossl_args = [
              'req',
              '-config', ossl.path,
              '-engine', 'pkcs11',
              '-new',
              '-key', @card_select.active_iter[0].to_s + ':45',
              '-keyform', 'engine',
              '-out', cert_file,
              '-text'
            ]
            ossl_args << '-x509' if @self_sign_check.active?

            @action = @self_sign_check.active? ? _("generating certificate") : _("generating sign request")
            system 'openssl', *ossl_args
          end
        end

        if @self_sign_check.active? or @upload_cert_check.active?
          pkcs_init_args = [
            '--reader', @card_select.active_iter[0].to_s,
            '--so-pin', @prefs.sopin,
            '--so-puk', '',
            '--pin', userpin,
            '--puk', '',
            '--store-certificate', cert_file,
            '--format', 'pem',
            '--auth-id', '02',
            '--id', '45'
          ]

          @action = _("storing certificate")
          system 'pkcs15-init', *pkcs_init_args
        end
        @action = ""
      end
    end
  end
end

